/*
	Copyright (C) 2010 Hans-Werner Hilse <hilse@web.de>
	See file "COPYING" for licence information.
*/
#include <stdlib.h>
#include <string.h>
#include "vipipc.h"

/* this configures how big the initial buffer is: */
#define IPCMESSAGE_PREALLOC_BUFFER 2048

extern ipcmessage ipcmessage_new() {
	ipcmessage msg = (ipcmessage) malloc(sizeof(ipcmessage_s));
	if(msg) {
		msg->data = malloc(IPCMESSAGE_PREALLOC_BUFFER);
		if(!msg->data) {
			free(msg);
			msg = NULL;
		} else {
			msg->data_bufsize = IPCMESSAGE_PREALLOC_BUFFER;
			msg->len = (int*) msg->data;
			*msg->len = sizeof(int);
			msg->ptr = msg->data + sizeof(int);
		}
	}
	return msg;
}

extern ipcmessage ipcmessage_read(char* data, const int length) {
	ipcmessage msg = (ipcmessage) malloc(sizeof(ipcmessage_s));
	if(msg) {
		msg->data = malloc(length);
		if(!msg->data) {
			free(msg);
			msg = NULL;
		} else {
			memcpy(msg->data, data, length);
			msg->data_bufsize = length;
			msg->len = (int*) msg->data;
			msg->ptr = msg->data + *msg->len;
		}
	}
	return msg;
}

extern void ipcmessage_free(ipcmessage* msg) {
	if(msg && *msg) {
		if((*msg)->data) {
			free((*msg)->data);
		}
		free(*msg);
		*msg = NULL;
	}
}

/* helpers for checking buffer boundaries: */

inline int check_msgdata_add(ipcmessage msg, int bytes) {
	if(!msg) return 1;
	if(msg->ptr + bytes > msg->data + msg->data_bufsize) {
		void *newbuf = realloc(msg->data,
			msg->data_bufsize + IPCMESSAGE_PREALLOC_BUFFER);
		if(!newbuf)
			return 1;
		msg->data_bufsize += IPCMESSAGE_PREALLOC_BUFFER;
		msg->data = newbuf;
	}
	return 0;
}

inline int check_msgdata_sub(ipcmessage msg, int bytes) {
	if(!msg) return 1;
	if(msg->ptr - bytes < msg->data)
		return 1;
	return 0;
}

/* push values onto the message stack */

extern int ipcmessage_push_int(ipcmessage msg, const int value) {
	if(check_msgdata_add(msg, sizeof(int)))
		return 1;

	*(int*)msg->ptr = value;
	msg->ptr += sizeof(int);
	*msg->len += sizeof(int);
	return 0;
}

extern int ipcmessage_push_byte(ipcmessage msg, const int value) {
	if(check_msgdata_add(msg, sizeof(char)))
		return 1;

	*(char*)msg->ptr = (char) value;
	msg->ptr += sizeof(char);
	*msg->len += sizeof(char);
	return 0;
}

extern int ipcmessage_push_string(ipcmessage msg, const char* buffer, const int length) {
	if(check_msgdata_add(msg, sizeof(int) + length))
		return 1;

	memcpy(msg->ptr, buffer, length);
	msg->ptr += length;

	*(int*)msg->ptr = length;
	msg->ptr += sizeof(int);

	*msg->len += sizeof(int) + length;
	return 0;
}

extern int ipcmessage_push_zstring(ipcmessage msg, const char* buffer) {
	return ipcmessage_push_string(msg, buffer, strlen(buffer)+1);
}


/* peek values from a message (i.e. do not move pointer) */

extern int ipcmessage_peek_int(ipcmessage msg, int* value) {
	if(check_msgdata_sub(msg, sizeof(int)))
		return 1;
	if(value)
		*value = *(int*)(msg->ptr - sizeof(int));
	return 0;
}

extern int ipcmessage_peek_byte(ipcmessage msg, int* value) {
	if(check_msgdata_sub(msg, sizeof(char)))
		return 1;
	if(value)
		*value = *(char*)(msg->ptr - sizeof(char));
	return 0;
}

extern int ipcmessage_peek_string(ipcmessage msg, char** buffer, int* length) {
	int thelength;
	if(ipcmessage_peek_int(msg, &thelength) || check_msgdata_sub(msg, thelength + sizeof(int)))
		return 1;
	if(buffer) {
		if(thelength) {
			*buffer = malloc(thelength);
			if(*buffer)
				memcpy(*buffer, msg->ptr - (thelength + sizeof(int)), thelength);
		} else {
			*buffer = NULL;
		}
	}
	if(length)
		*length = thelength;
	return 0;
}

/* pop values from a message (i.e. move pointer) */

extern int ipcmessage_pop_string(ipcmessage msg, char** buffer, int* length) {
	int thelength;
	if(ipcmessage_peek_string(msg, buffer, &thelength))
		return 1;
	msg->ptr -= thelength + sizeof(int);
	if(length)
		*length = thelength;
	return 0;
}

extern int ipcmessage_pop_byte(ipcmessage msg, int* value) {
	if(ipcmessage_peek_byte(msg, value))
		return 1;
	msg->ptr -= sizeof(char);
	return 0;
}

extern int ipcmessage_pop_int(ipcmessage msg, int* value) {
	if(ipcmessage_peek_int(msg, value))
		return 1;
	msg->ptr -= sizeof(int);
	return 0;
}

