/*
	Copyright (C) 2010 Hans-Werner Hilse <hilse@web.de>
	See file "COPYING" for licence information.
*/

/***************************************************************************
* test the libvipipcstack
*
* creates a program that implements either a server or a client for
* IPC, depending on the first argument ("receiver" / "sender")
*
* Client reads from stdin and sends to server, which outputs on stdout.
* Kind of pointless, but it shows if servers/clients are working as
* expected.
*
* Server uses select(), which is a bit overkill for a server that will
* only have one client at a time.
*
* see progs/example-informationservice.c for a better documented
* example using the VIP1710 informationservice.
***************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "../vipipc.h"
#include "../vipipc_prototypes.h"

#define TESTSOCKET "/tmp/ipcpipe#0"

/* interface definitions: */
IPCPROTO_NEW_SERVER(MyTestService, "IfaceTestService")
IPCPROTO_NEW_PARSER(MyTestService, DoComm, "IS")
IPCPROTO_NEW_DISPATCHER(MyTestService, DoComm)

IPCPROTO_NEW_GENERIC(MyTestServiceCaller, "IfaceTestService")
IPCPROTO_NEW_CALL(MyTestServiceCaller, DoComm, "#P|E")

int ms_docomm(ipcmessage msg, void* data) {
	char *stringval;
	int stringlength;
	int err;
	ipcconn client = (ipcconn) data;

	err = MyTestService_parse_DoComm(msg, &stringlength, &stringval);
	if(err) {
		ipcconn_procedure(client, NULL, "iSb!", err, "Some error occurred.", 1);
	} else {
		write(1, stringval, stringlength);
		free(stringval);
		ipcconn_procedure(client, NULL, "b!", 0);
	}
	return 0;
}

int main(int argc, char* argv[]) {
	int err;

	int nfds;
	fd_set rfds;
	struct timeval tv;

	ipcconn ms, ms_client, ms_caller;

	if(argc < 2) {
		fprintf(stderr, "error: invoke with either \"receiver\" or \"sender\" as first argument!\n");
		exit(1);
	}

	if(!strcmp("sender", argv[1])) {
		/* client implementation: */
		ipcconn ms_caller;
		char *buf = malloc(1024);
		if(!buf) {
			fprintf(stderr, "error creating buffer\n");
			exit(1);
		}

		ms_caller = MyTestServiceCaller_new(TESTSOCKET);
		if(!ms_caller) {
			fprintf(stderr, "error creating ms_caller ipcconn\n");
			exit(1);
		}
		do {
			err = read(0, buf, 1024);
			if(err == -1) {
				perror("read()");
			} else {
				err = MyTestServiceCaller_DoComm(ms_caller, buf, err);
				if(err)
					fprintf(stderr, "error calling DoComm procedure.\n");
			}
		} while(!err);

		ipcconn_free(&ms_caller);
		
		exit(0);
	} else if(!strcmp("receiver", argv[1])) {
		/* server implementation: */
		int ret;

		ms = MyTestService_new(TESTSOCKET);
		if(!ms) {
			fprintf(stderr, "error creating ms ipcconn\n");
			exit(1);
		}

		err = ipcconn_bind_and_listen(ms, 0777, 1);
		if(err) {
			fprintf(stderr, "error listening+binding to socket\n");
			ipcconn_free(&ms);
			exit(1);
		}
	
		ms_client = NULL;

		/* for now, we'll run forever (or 'til you hit CTRL-C) */
		while(1) {
			FD_ZERO(&rfds);
			FD_SET(ms->fd, &rfds);
			if(ms_client) {
				FD_SET(ms_client->fd, &rfds);
				nfds = (ms->fd > ms_client->fd) ? ms->fd + 1: ms_client->fd + 1;
			} else {
				nfds = ms->fd + 1;
			}

			tv.tv_sec = 10;
			tv.tv_usec = 0;

			ret = select(nfds, &rfds, NULL, NULL, &tv);

			if(ret == -1) {
				perror("select()");
				break;
			} else if(ret) {
				if(ms_client && FD_ISSET(ms_client->fd, &rfds)) {
					char *procedure;
					char *interface;
					int status;
					ipcmessage msg;
					err = ipcconn_receive_rpc(ms_client,
							&msg, &status, &interface, &procedure);
					if(!err) {
						if(MyTestService_dispatch_DoComm(interface, procedure, msg,
							ms_docomm, (void*) ms_client) < 0)
							fprintf(stderr,
								"undispatched procedure call: %s:%s\n",
								interface, procedure);

						if(msg) ipcmessage_free(&msg);
						free(procedure);
						free(interface);
					} else {
						ipcconn_free(&ms_client);
					}
				}
				if(FD_ISSET(ms->fd, &rfds)) {
					/* just disconnect existing client session */
					if(ms_client) ipcconn_free(&ms_client);
					ipcconn_accept(ms, &ms_client);
				}
			}
		}
	}
}

